--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2010 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author:
--------------------------------------------------------------------------------
-- Stefan Fischer <Stefan.Fischer@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH CESYS PRODUCTS ONLY!
-- YOU ARE NOT ALLOWED TO MODIFY AND DISTRIBUTE OR USE IT WITH ANY OTHER
-- HARDWARE, SOFTWARE OR ANY OTHER KIND OF ASIC OR PROGRAMMABLE LOGIC DESIGN
-- WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;


entity flash_ctrl is
  port
  (
    rst : in std_logic;
    clk : in std_logic;
    
    usr_access_i : in std_logic;
    usr_erase_i : in std_logic;
    usr_we_i : in std_logic;
    usr_wr_adr_i : in std_logic_vector(24 downto 0);
    usr_dati_i : in std_logic_vector(31 downto 0);

    usr_rd_adr_i : in std_logic_vector(24 downto 0);
    usr_dato_o : out std_logic_vector(31 downto 0);
    write_ok_o : out std_logic;
    read_ok_o  : out std_logic;

    flash_rdy_o  : out std_logic;
    flash_s_n_o : out std_logic;
    flash_c_o : out std_logic;
    flash_hold_n_o : out std_logic;
    flash_w_n_o : out std_logic;
    flash_d_o : out std_logic;
    flash_q_i : in std_logic
  );
end flash_ctrl;


architecture RTL of flash_ctrl is

  signal usr_access : std_logic := '0';
  signal usr_erase  : std_logic := '0';
  signal usr_high_addr  : std_logic := '0';
  signal s_write_ext_addr  : std_logic := '0';
  signal upper_segment  : std_logic := '0';
  signal usr_we     : std_logic := '0';
  signal s_write_pending : std_logic := '0';
  signal s_read_pending  : std_logic := '0';
  signal usr_rd_adr : std_logic_vector(23 downto 0) := (others  => '0');
  signal usr_wr_adr : std_logic_vector(23 downto 0) := (others  => '0');
  signal usr_dati   : std_logic_vector(31 downto 0) := (others  => '0');
  signal usr_dato   : std_logic_vector(31 downto 0) := (others  => '0');
  signal write_ok     : std_logic := '0';
  signal read_ok      : std_logic := '0';

  signal flash_s_n : std_logic := '1';
  signal flash_c : std_logic := '0';
  signal flash_hold_n : std_logic := '1';
  signal flash_w_n : std_logic := '1';
  signal flash_d : std_logic := '0';
  signal flash_q : std_logic := '0';

  signal shiftregister_in : std_logic_vector(7 downto 0) := (others => '1');
  signal shiftregister_out : std_logic_vector(7 downto 0) := (others => '0');
  signal clk_divider : std_logic := '0';
  signal is_first_run : std_logic := '1';
  signal cyc_count : unsigned(6 downto 0) := (others => '0');

  constant WREN : std_logic_vector(7 downto 0) := x"06";
  constant WEAR : std_logic_vector(7 downto 0) := x"C5";  -- WRITE EXTENDED ADDRESS REGISTER
  constant BE : std_logic_vector(7 downto 0) := x"C7";
  constant SSE : std_logic_vector(7 downto 0) := x"20";
  constant SE : std_logic_vector(7 downto 0) := x"D8";
  constant PP : std_logic_vector(7 downto 0) := x"02";
  constant RDSR : std_logic_vector(7 downto 0) := x"05";
  constant FAST_READ : std_logic_vector(7 downto 0) := x"0B";

  type t_states is
  (
    s_STARTUP,
    S_IDLE,
    S_WRITE_EN,
    S_WR_EXT_ADD_REG,
    S_SECTOR_ERASE,
    S_PAGE_PROG,
    S_RDSR_COM,
    S_RDSR,
    S_RDSR_SH,
    S_FAST_READ
  );
  signal state : t_states := s_STARTUP;

  signal databuffer : std_logic_vector(23 downto 0) := (others => '0');

begin
  flash_rdy_o <= '1' when state = s_IDLE else '0';
  usr_access <= usr_access_i;
  usr_erase <= usr_erase_i;
  usr_we <= usr_we_i;
  usr_high_addr <= ( (usr_wr_adr_i(24) and usr_we) or (usr_rd_adr_i(24) and (not usr_we)) );

  usr_wr_adr <= usr_wr_adr_i(23 downto 0);
  usr_dati <= usr_dati_i;

  usr_rd_adr <= usr_rd_adr_i(23 downto 0);
  usr_dato_o <= usr_dato;

  write_ok_o <= write_ok;
  read_ok_o <= read_ok;

  flash_s_n_o <= flash_s_n;
  flash_c_o <= flash_c;
  flash_hold_n_o <= flash_hold_n;
  flash_w_n_o <= flash_w_n;
  flash_d_o <= flash_d;
  flash_q <= flash_q_i;

  usr_dato <= shiftregister_in & databuffer;

  flash_hold_n <= '1';
  flash_w_n <= '1';

  flash_c <= clk_divider and (not flash_s_n);
  flash_d <= shiftregister_out(7);

  process(clk)
  begin
    if clk'event and clk = '1' then
      if rst = '0' then
        state <= s_STARTUP;
        clk_divider <= '0';
        cyc_count <= (others => '0');
      else

        write_ok <= '0';
        read_ok <= '0';
        clk_divider <= not clk_divider;
        if clk_divider = '1' then
          cyc_count <= cyc_count + 1;
          shiftregister_out <= shiftregister_out(6 downto 0) & '0';
          shiftregister_in <= shiftregister_in(6 downto 0) & flash_q;
        end if;
        
        case state is
          when s_STARTUP =>
            -- the FPGA needs 3 clock cycles to put userclk on cclk
            -- let's wait for 3 more clock cycles (total 6)
            -- Before we write a byte, let's deassert chip select
            flash_s_n <= '0';
            if cyc_count = 5 then
              flash_s_n <= '1';
              cyc_count <= (others => '0');
              clk_divider <= '0';
              shiftregister_out <= (others => '0');
              state <= s_IDLE;
            end if;

          when S_IDLE =>
            clk_divider <= '0';
            cyc_count <= (others => '0');
            if usr_access = '1' then
              flash_s_n <= '0';

              if ( (usr_high_addr xor upper_segment) = '1') then
                -- LSB of the EXTENDED ADDRESS REGISTER
                s_write_ext_addr  <= '1';
                -- clear all pending requests at write_ok
                if usr_we = '1' then
                  s_write_pending <= '1';
                else
                  s_read_pending  <= '1';
                end if;
                shiftregister_out <= WREN;
                state <= S_WRITE_EN;
              else
                s_write_ext_addr  <= '0';
                if usr_we = '1' or s_write_pending = '1' then
                  shiftregister_out <= WREN;
                  state <= S_WRITE_EN;
                else
                  shiftregister_out <= FAST_READ;
                  state <= S_FAST_READ;
                end if;
              end if;
            end if;

          when S_WRITE_EN =>
            if clk_divider = '1' then
              if cyc_count = 10 then
                flash_s_n <= '0';
                cyc_count <= (others => '0');
                if s_write_ext_addr = '1' then
                  shiftregister_out <= WEAR;
                  state <= S_WR_EXT_ADD_REG;
                else
                  if usr_erase = '1' then
                    s_write_pending <= '0';
                    shiftregister_out <= SE;
                    state <= S_SECTOR_ERASE;
                  else
                    s_write_pending <= '0';
                    shiftregister_out <= PP;
                    state <= S_PAGE_PROG;
                  end if;
                end if;
              end if;

            elsif cyc_count = 8 then
              flash_s_n <= '1';
            end if;
          
          when S_WR_EXT_ADD_REG =>
            if clk_divider = '1' then
              if cyc_count = 7 then

                shiftregister_out <= "0000000" & not (upper_segment);
                upper_segment <= not upper_segment;

              elsif cyc_count = 18 then
                s_write_ext_addr  <= '0';
                cyc_count <= (others => '0');
                -- state <= s_IDLE;
                flash_s_n <= '0';
                if s_write_pending = '1' then
                  shiftregister_out <= WREN;
                  state <= S_WRITE_EN;
                else
                  shiftregister_out <= FAST_READ;
                  state <= S_FAST_READ;
                end if;
              end if;

            elsif cyc_count = 16 then
              flash_s_n <= '1';
            end if;

          when S_SECTOR_ERASE =>
            if clk_divider = '1' then
              if cyc_count = 7 then
                shiftregister_out <= usr_wr_adr(23 downto 16);
              elsif cyc_count = 15 then
                shiftregister_out <= usr_wr_adr(15 downto 8);
              elsif cyc_count = 23 then
                shiftregister_out <= usr_wr_adr(7 downto 0);
              elsif cyc_count = 34 then
                flash_s_n <= '0';
                cyc_count <= (others => '0');
                shiftregister_out <= RDSR;
                state <= S_RDSR_COM;
              end if;
            elsif cyc_count = 32 then
              flash_s_n <= '1';
            end if;
          -- end if;

          when S_PAGE_PROG =>
            if clk_divider = '1' then
              if cyc_count = 7 then
                shiftregister_out <= usr_wr_adr(23 downto 16);
              elsif cyc_count = 15 then
                shiftregister_out <= usr_wr_adr(15 downto 8);
              elsif cyc_count = 23 then
                shiftregister_out <= usr_wr_adr(7 downto 0);
              elsif cyc_count = 31 then
                shiftregister_out <= usr_dati(7 downto 0);
              elsif cyc_count = 39 then
                shiftregister_out <= usr_dati(15 downto 8);
              elsif cyc_count = 47 then
                shiftregister_out <= usr_dati(23 downto 16);
              elsif cyc_count = 55 then
                shiftregister_out <= usr_dati(31 downto 24);
              elsif cyc_count = 66 then
                flash_s_n <= '0';
                cyc_count <= (others => '0');
                shiftregister_out <= RDSR;
                state <= S_RDSR_COM;
              end if;
            elsif cyc_count = 64 then
              flash_s_n <= '1';
            end if;

          when S_RDSR_COM =>
            if clk_divider = '1' then
              if cyc_count = 15 then
                cyc_count <= (others => '0');
                state <= S_RDSR;
              end if;
            end if;

          when S_RDSR =>
            if clk_divider = '1' then
              if cyc_count = 7 then
                cyc_count <= (others => '0');
              end if;
            else
              if (cyc_count = 0) and shiftregister_in(0) = '0' then
                write_ok <= '1';
                clk_divider <= '0';
                flash_s_n <= '1';
                state <= S_RDSR_SH;
              end if;
            end if;

          when S_RDSR_SH =>
            if clk_divider = '1' then
              if cyc_count = 1 then
                cyc_count <= (others => '0');
                state <= S_IDLE;
              end if;
            end if;

          when S_FAST_READ =>
            if clk_divider = '1' then
              if cyc_count = 7 then
                shiftregister_out <= usr_rd_adr(23 downto 16);
              elsif cyc_count = 15 then
                shiftregister_out <= usr_rd_adr(15 downto 8);
              elsif cyc_count = 23 then
                shiftregister_out <= usr_rd_adr(7 downto 0);
              elsif cyc_count = 31 then -- 8 DUMMY CYCLES
                shiftregister_out <= (others => '0');
              elsif cyc_count = 48 then
                databuffer(7 downto 0) <= shiftregister_in;
              elsif cyc_count = 56 then
                databuffer(15 downto 8) <= shiftregister_in;
              elsif cyc_count = 64 then
                databuffer(23 downto 16) <= shiftregister_in;
              elsif cyc_count = 71 then
                read_ok <= '1';
              end if;
            else
              if cyc_count = 72 then
                flash_s_n <= '1';
              elsif cyc_count = 74 then
                cyc_count <= (others => '0');
                state <= S_IDLE;
              end if;

          end if;

          when others => null;

        end case;
        
      end if; -- reset
    end if; -- clk
  end process;

end RTL;
